////////////////////////////////////////////////////////////////////////////////
//                                  Include                                   //
////////////////////////////////////////////////////////////////////////////////


#include "Oracle.h"
#include <fstream.h>

////////////////////////////////////////////////////////////////////////////////
//                              Public: Oracle                                //
////////////////////////////////////////////////////////////////////////////////


Oracle::Oracle(char* filename) : x(env), u(env), aI(env), aJ(env), cx(env), gu(env),
				 obj(env), model(env), cplex(env)
{

  try {
    //----- User: Create data model -------------------------------------------1

    // Initiate variables
    IloInt i, j, aCount, aCol;


    // Data file
    ifstream file(filename);

    // m (rows), n (cols)
    file >> m >> n;

    // x (primal variables)
    x.add(IloNumVarArray(env, n, 0, 1));

    // u (dual multipliers)
    u.add(IloNumArray(env, m));
    // c (cost of building j)
    IloNumArray c(env, n);
    for (j = 0; j < n; j++)
      file >> c[j];

    IloNum sparsity;
    sparsity = 0.1;
    
    aI.add(IloIntArray(env, m+1));

    IloInt Icur;
    Icur = 0;
    // a (cover matrix)

    for (i = 0; i < m; i++) {
      // a[i] = IloBoolArray(env, n);
      aI[i] = Icur;
      file >> aCount;
      aJ.add(IloIntArray(env, aCount));

      for (j = 0; j < aCount; j++) {
        file >> aCol;
        // a[i][aCol - 1] = 1;
	aJ[Icur] = aCol -1;
	//	cout << "Icur="<< Icur  << "ok\n";
	Icur++;
      }
    }
    aI[m] = Icur;
 
    // model: object
    for (j = 0; j < n; j++)
      cx += c[j] * x[j];

    // model: lagrange relaxad cover constraints
    IloExpr lag(env);

    // model: minimize
    obj = IloMinimize(env, cx + lag);
    model.add(obj);

    // model: x constraints
    for (j = 0; j < n; j++) {
      model.add(x[j] >= 0);
      model.add(x[j] <= 1);
    }

    // cplex: extract model
    cplex.extract(model);

    //-------------------------------------------------------------------------1

    //----- User: Variables that must be set ----------------------------------2

    // u size
    uSize = m;

    // x dimensions and size
    xDims = 1;
    xDim1Size = n;

    // Add all unconstrained multipliers u, end with an InINF
    ucVector = new unsigned int[1];
    ucVector[0] = InINF;

    // Is the primal problem of min type
    isMinProblem = true;

    //-------------------------------------------------------------------------2
	
  } catch(IloException& e) {
    cerr << "Error: " << e.getMessage() << endl;
  } catch(...) {
    cout << "Error" << endl;
  }
}


////////////////////////////////////////////////////////////////////////////////
//                              Public: ~Oracle                               //
////////////////////////////////////////////////////////////////////////////////


Oracle::~Oracle()
{
  env.end();
}


////////////////////////////////////////////////////////////////////////////////
//                               Public: Solve                                //
////////////////////////////////////////////////////////////////////////////////


void Oracle::Solve(const double uVector[], double &fValue, double sgVector[],
		   double xVector[])
{
  //----- User: Solve model ----------------------------------------------------

  // Inititate variables
  IloInt i, j;
  IloNum sum;

  // Update u
  u.clear();
  for (i = 0; i < m; i++)
    u.add(uVector[i]);

  // Update model
  model.remove(obj);

  IloExpr lag(env);
  
  for (i = 0; i < m; i++) {
    IloExpr temp(env);
    for (j = aI[i]; j < aI[i+1]; j++)
      temp +=  x[aJ[j]];
    //      temp += a[i][j] * x[j];
    lag += u[i] * (1 - temp);
    temp.end();
  }

  obj = IloMinimize(env, cx + lag);
  lag.end();

  model.add(obj);

  // Solve model
  cplex.solve();
  
  // fValue (with "-" beacuse min problem)
  fValue = -cplex.getValue(obj);

  // sgVector (with "-" beacuse min problem)
  for (i = 0; i < m; i++) {
    sum = 0;
    for (j = aI[i]; j < aI[i+1]; j++)
      sum +=  cplex.getValue(x[aJ[j]]);
    sgVector[i] = -(1 - sum);
  }

  // xVector
  for (j = 0; j < n; j++)
    xVector[j] = cplex.getValue(x[j]);
}


////////////////////////////////////////////////////////////////////////////////
//                                Public: getX                                //
////////////////////////////////////////////////////////////////////////////////


long Oracle::getUSize()
{
  return uSize;
}

long Oracle::getXDims()
{
  return xDims;
}

long Oracle::getXDim1Size()
{
  return xDim1Size;
}

long Oracle::getXDim2Size()
{
  return xDim2Size;
}

bool Oracle::getIsMinProblem()
{
  return isMinProblem;
}

unsigned int* Oracle::getUCVector()
{
  return ucVector;
}


////////////////////////////////////////////////////////////////////////////////
