/*	
set covering problem (scp) file format convert
by douglas potter 2008

this code takes one of the following two input file types
and generates one or more of the following two output
file types:
inputs: 1.)standard scp file 2.) rail scp file
outputs: 1.) standard scp file 2.) matlab .mat file

the program's construction is an overkill for the tasks performed to ensure
that new features can be easily added.

design priorities (in order of importance):
1. use std lib to reduce machine dependencies
2. readibility
3. speed
4. memory usage

almost everything is in good shape, except perhaps some of the 
file IO error handling could be improved

the code supplied with extra superfluous comments for
pedagogical reasons

functions in program:
main - entry point for program
usage - displays usage text
*/

#include <iostream>
#include <fstream> 
#include <string>
#include <list>
#include <vector>
#include <sstream>
#include <map>
#include "mat.h"
#include "matrix.h"

/* explanation of the head files:
header file 	- functionality added

iostream 	- standard input output
fstream 	- input output for files
string 		- support C++ strings
list		- support lists
vector 		- support vectors
map		- support for maps and multimaps
sstream		- support string streams, used for converting chars in files to ints
mat.h		- support matlab mat files
matrix.h	- support matlab matrices
*/

using namespace std; /* now all standard (std) library
classes and functions are accessible without
a scope operator. the means that i
can type 'cout' instead of 'std:cout' for example in main
*/

// input and outputs could probably be combined
enum inputs {iStdScp=1, iRailScp=2}; // input file types
enum outputs {oStdScp=1, oMat=3}; // output file types

enum parseError {numArg=1, help, nothingToDo, syntaxArg, tooManyInput, tooManyOutput,
		errorMatch}; // errors while parse arguments
				
struct task { int type; string fileName; }; // define task structure

// functions
void usage(); // display text describing usage & syntax
int parse(const int &argc,  char *const argv[], list<task> &input, list<task> &output);
int parseError(int error);


/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
		/* entry point for the program when run */
main(int argc, char *argv[]) 
/* argc = number of command-line arguments, *argv[] pointers to the arguments (aka args)*/
{
	list<task> input, output; // lists of input and output tasks	
	/* use vector in order to let C++ take care of
		memory management */	
	int dimM,dimN; // dims of cover matrix
	vector<int> costVector; // cost vector for the scp	
	vector<int> col; // for storing a col	
	multimap<int,int> xyPair; // for storing coordinate of nonzero covering matrix entries
	multimap<int,int> yxPair; // for storing coordinate of nonzero covering matrix entries
				  // whether xyPair or yxPair is used depends on direction of conversion
	typedef multimap<int,int>::iterator pairIter; 
	int numNon=0; // number of nonzero entries		
													
	int error = parse(argc, argv, input, output); // parse arguments
	
	if (error){ // handle error from parser
		parseError(error); // display somewhat instructive error text
		return -1; // exit program
	}
	
	fstream file; // file object for the input and output files
	
	// check that files given are valid.
	
	/* 	the code for checking input file(s) is excessive since
		only one input file is allowed. but maybe more input
		files in the future??
	*/
	for(list<task>::iterator iter = input.begin(); 
				iter != input.end(); ++iter){
		file.open(iter->fileName.c_str());
		if(!file){
			cerr << "\n Error: Can't open input file: " <<
				iter->fileName << endl;
			return -1;
		}
		file.close();
		file.clear();
	}
	
	// make sure output files do NOT exist
	for(list<task>::iterator iter = output.begin();
				iter != output.end(); ++iter){
		file.open(iter->fileName.c_str());
		if(file){
			cerr << "\nError: Problem with output file: " <<
				 iter->fileName << "\n" << 
				"Output files may NOT exist before the convscp is run.\n" <<
				 endl;
			return -1;
		}
		file.close();
		file.clear();		
	}
	
	// the following code was designed for one input file	
	list<task>::iterator inputIter = input.begin();
	file.open(inputIter->fileName.c_str(), fstream::in); // open input file as input
	
	// get the dim of the set cover matrix 
	char fileBuf[2048]; // if a single has more than 2k of characters the program may crash
	file.getline(fileBuf, 1024); // get the first line
	stringstream streamBuf(fileBuf); // put the 1st line into a string stream

	streamBuf >> dimM >> dimN; // read the covering matrix dimensions
	
	if(dimM < 1 || dimN < 1){ // are dimensions reasonable?
		cerr << "Error: The dimensions of the covering matrix are invalid." <<
			"\nCheck input file." << endl;	
		return -1;
	}
	
	
	costVector.clear(); // clear the costVector for safety
	costVector.resize(dimN); // make costVector big enough
	int numEle=0; // number of elements 
	int j; // iterator over elements
	int iterCol=0; // iterator for columns
	int iterRow=0; // iterator for rows
	int tmp;

	switch(inputIter->type){
	case iStdScp: // process standard scp input file
		
		// read the costs
		for(iterCol=0;iterCol<dimN;++iterCol){ // save rows that the column covers
			if( !(streamBuf >> costVector[iterCol]) ){ // current line exhausted
				streamBuf.clear(); 
				file.getline(fileBuf, 2048); // proceed to next line
				streamBuf.str(fileBuf); // put line into a string stream
				streamBuf >> costVector[iterCol];
			 }			 	
		}
		
		numNon = 0; // set number of nonzero elements to zero
		iterRow = 0; 
		while(file && iterRow < dimM ){ 
			if( !(streamBuf >> numEle) ){ // current line exhausted
				streamBuf.clear(); 
				file.getline(fileBuf, 2048); // proceed to next line
				streamBuf.str(fileBuf); // put line into a string stream
				streamBuf >> numEle;
			 }	
					
			iterRow++; // row counter + 1
			
			for(j=0;j<numEle;++j,++numNon){ // save rows that the column covers
			 	if( !(streamBuf >> tmp) ){ // current line exhausted
			 		streamBuf.clear(); 
			 		file.getline(fileBuf, 2048); // proceed to next line
			 		streamBuf.str(fileBuf); // put line into a string stream
			 		streamBuf >> tmp; // throw away duplicate element
			 	}
			 	xyPair.insert(make_pair(iterRow,tmp)); 				 	
			}		 			
		}
							
		#ifndef NDEBUG
		cout << "\n\n standard scp input info\n" <<
			"\nxyPair.size() = " << xyPair.size() <<
			"\niterRow=" << iterRow << "  dimM=" << dimM << " dimN=" << dimN << endl;
		#endif
		
		if(dimM != iterRow){
			cerr << "Error: file did NOT contain the number of " << 
				"rows specified in the file or some " <<
				"problem reading the input file" << endl;
			return -1;
		}								
	
			
		break;
	case iRailScp:	 // process rail format scp input file
		numNon = 0; // set number of nonzero elements to zero
		iterCol = 0; 
		while(file && iterCol < dimN){ 
			if( !(streamBuf >> costVector[iterCol]) ){ // current line exhausted
				streamBuf.clear(); 
				file.getline(fileBuf, 2048); // proceed to next line
				streamBuf.str(fileBuf); // put line into a string stream
				streamBuf >> costVector[iterCol];
			 }	
		
			streamBuf >> numEle; // how many rows does this column cover
			
			iterCol++; // column counter + 1
			
			for(j=0;j<numEle;++j,++numNon){ // save rows that the column covers
				streamBuf >> tmp;
				xyPair.insert(make_pair(tmp,iterCol));
			 	if(!streamBuf){ // current line exhausted
			 		streamBuf.clear(); 
			 		file.getline(fileBuf, 2048); // proceed to next line
			 		streamBuf.str(fileBuf); // put line into a string stream
			 		streamBuf >> tmp; // throw away duplicate element
			 	}				 	 	
			}
		}				
	
		#ifndef NDEBUG
		cout << "\n\n rail input info\n" <<
			"\nxyPair.size() = " << xyPair.size() <<
			"\niterCol=" << iterCol << "  dimN=" << dimN << endl;
		#endif
	
		if(dimN != iterCol){
			cerr << "Error: file did NOT contain the number of " << 
				"columns specified in the file or other" <<
				"problem reading the input file" << endl;
			return -1;
		}
																																																											
		break;
		
	default: // this shouldn't happen, the parser should catch this.
		cerr << "Error: Undefined input type." << endl;	
		return -1;
		break;
	}		
				
	file.close(); // close input file
	file.clear(); /* clear any file errors, so they don't effect the output file
			could add additional error handeling here*/
	
	
	// process output files
	for(list<task>::iterator iterOutputTask = output.begin();
				iterOutputTask != output.end();
				++iterOutputTask) {
		
		switch(iterOutputTask->type){
		
		case oStdScp:
		
  			if(xyPair.size() == 0){ // if xyPair is empty, populate it
				for(multimap<int,int>::iterator pairIter = yxPair.begin(); 
					pairIter != yxPair.end(); ++pairIter){
					xyPair.insert(make_pair(pairIter->second,pairIter->first));
				}
			 	yxPair.clear(); // clear yxPair to save memory
		   	}
		   					
			file.open(iterOutputTask->fileName.c_str(), fstream::out); // open output file as output
			file << dimM << " " << dimN;				
			
			iterCol = 0; // used for formating 
			for(vector<int>::iterator iter = costVector.begin();
						iter != costVector.end(); ++iter, ++iterCol){
				if( !(iterCol % 15) ){ 
						file << "\n";
					}else{
						file << " ";
					}
				file << *iter;						
			}				
			
			for(iterRow=1;iterRow <= dimM;++iterRow){
				file << "\n" << xyPair.count(iterRow);
				j=0; // counter for line breaks
				pair<pairIter, pairIter> pos = xyPair.equal_range(iterRow); // which cols in this row
				while(pos.first != pos.second){								
					if( !(j % 15) ){ 
						file << "\n";
					}else{
						file << " ";
					}
					file << pos.first->second;
					pos.first++;
					j++;
				}				
			}				
						
			file.close();
			file.clear();
			break;
			
		case oMat: /* this section relies heavily on C and the matlab header files
				for creating mat files */
		
			{ // define a code block to define vars inside of the switch 
			
			MATFile *matFile; // create matlab mat-file handle
			matFile = matOpen(iterOutputTask->fileName.c_str(), "w"); // assign file to handle
			
			if (matFile == NULL) {
   				 cerr << "\nError: Matlab(R) Mat output file could not be created." << endl;
   				 return -1;
			}					
			
			mwSize matM = dimM;
			mwSize matN = dimN;
			mwSize nzmax = numNon+1;
			
			mxArray *matCover; 
			mxArray *matCost;	
			double *si,*sr;	// for matlab sparse matrix			
			mwIndex *jcs,*irs; //  for matlab sparse matrix
			mwIndex sparseIndex;				
																											
			matCost = mxCreateDoubleMatrix(1,matN,mxREAL); // allocate matlab dense matrix			
			sr = mxGetPr(matCost);
		
			for(vector<int>::iterator iter = costVector.begin(); // copy costVector
						iter != costVector.end(); ++iter) *(sr++) = *iter;	
			
			
			if(matPutVariable(matFile, "cost", matCost)){
				cerr << "\nError: Could not save cost vector to Matlab(R) Mat output file" 
					<< endl;
				return -1;
			}
					
			mxDestroyArray(matCost); // free matCost	
			
			matCover = mxCreateSparse(matM, matN, nzmax, mxREAL); // allocate matlab sparse matrix
			
			if (matCover == NULL) {
      				cerr << "\nErro: Could NOT create Matlab(R) Mat cover matrix." << endl;
      				return -1;
      			}
      			
      			if(yxPair.size() == 0){ // if yxPair is empty, populate it
				for(multimap<int,int>::iterator pairIter = xyPair.begin(); 
					pairIter != xyPair.end(); ++pairIter){
					yxPair.insert(make_pair(pairIter->second,pairIter->first));
				}
			 	xyPair.clear(); // clear xyPair to save memory
		   	}
      			
      			sr  = mxGetPr(matCover);
  			// si  = mxGetPi(matCover); no complex elements, so not needed
  			irs = mxGetIr(matCover);
  			jcs = mxGetJc(matCover);

 			
			sparseIndex = 0;
      			for(iterCol=0;iterCol < dimN; ++iterCol){
				pair<pairIter, pairIter> pos = yxPair.equal_range(iterCol+1);//which rows in this col
				jcs[iterCol]=sparseIndex; // potential problem: indexing jcs with an int
				while(pos.first != pos.second){					
				 	sr[sparseIndex] = 1; // value of nonzero element
				 	irs[sparseIndex] = pos.first->second - 1; // save x coor
					pos.first++;
					sparseIndex++;
				}				
			}					
			jcs[iterCol]=sparseIndex;
			
  			if(matPutVariable(matFile, "cover", matCover)){
				cerr << "\nError: Could not save cover matrix to Matlab(R) Mat output file" 
					<< endl;
				return -1;
			}		
		
			mxDestroyArray(matCover);
			
			if (matClose(matFile)) {
    				cerr << "\nError: Could NOT close Matlab(R) Mat file." << endl;
      				return -1;
      			} 
			
			} // end of switch code block
			break;
				
		default: // this shouldn't happen, the parser should catch this.
			cerr << "Error: Undefined output type." << endl;	
			return -1;
			break;
		
		}	
	}
	
	#ifndef NDEBUG
	cerr << "\nInput file types and file names:" << endl;
	for(list<task>::iterator iter = input.begin();
				iter != input.end();
				++iter) {
		cout << iter->type <<  " " << iter->fileName <<endl;
	}
	
	cerr << "\nOutput file types and file names:" << endl;
	for(list<task>::iterator iter = output.begin();
				iter != output.end();
				++iter) {
		cout << iter->type <<  " " << iter->fileName <<endl;
	}
	#endif				

	return 0; // exiting program sucessfully
}

void usage()
{
	cerr << "convscp converts setting covering problem (scp) formats\n\n" <<
		"usage: convscp -i(type) input_file -o(type) output_file\n\n" <<
		"Only one input is allowed. One to two outputs are allowed; the outputs " <<
		"cannot be of the same type as each other nor the input. \n\n" <<
		"valid options for -i: s standard scp format, r rail scp format\n" <<
		"valid options for -o: s standard scp format, m matlab(R) mat format\n\n" <<
		"Examples of valid syntax: \n" <<
		"convscp -ir rail507 -os rail507standard\n" <<
		"convscp -ir rail -os railstd -om railmat\n" <<
		"convscp -is scp41.txt -om scp41.mat\n" << 
		"convscp -om scp41.mat -is scp41.txt\n\n" << 
		"Note 1: the syntax is case senstive\n" <<
		"Note 2: converting large files can take a long time\n" << 
		"Note 3: masks or wildcards, ie for selecting multiple files, are not allowed\n" << 
		"Note 4: the end of the file name (ie the file extension) does not effect the output\n" <<
		"Note 5: input file errors may crash convscp and may result in corrupt output files\n" <<  
		 endl;

}
		
int parse(const int &argc,  char *const argv[], list<task> &input, list<task> &output)
{	
	if(argc <= 1) return numArg; /* catch simple case 1st, to make sure
				     'string cmdArg[argc];' doesn't seg fault */
	string cmdArg[argc]; // create c++ array of strings for the args

	for(int a=0; a < argc; ++a) cmdArg[a].assign(argv[a]); /* put
			c-style strings *argv[] into c++ string cmdArg[] */

	#ifndef NDEBUG /* if compiled w/'-DNDEBUG' option
	  the code between #ifndef and #endif won't be compiled 
	  DNDEBUG = Do Not DEBUG */
		cout << argc << " arguments supplied:";
		for(int a=0; a< argc; ++a) cout << "\n" << a + 1 << ".) "<< cmdArg[a];
		cout << endl;	
	#endif

	if(argc != 3 && argc != 5 && argc != 7){ // check number of arguments
		if(cmdArg[1] == "--help"){ 		
			return help;
		}else{
			return numArg;
		}
	}
	
	task parsedTask; // for constructing input and output tasks

	for(int a=1; a < argc; a+=2)
	{
		if( cmdArg[a] == "-is") // rail scp format input
		{
			parsedTask.type = iStdScp;
			parsedTask.fileName = cmdArg[a+1];
			input.push_back(parsedTask);		
		} 
		else if( cmdArg[a] == "-ir") // standard scp format input
		{
			parsedTask.type = iRailScp;
			parsedTask.fileName = cmdArg[a+1];
			input.push_back(parsedTask);		
		}
		else if( cmdArg[a] == "-os" ) // standard scp format output
		{
			parsedTask.type = oStdScp;
			parsedTask.fileName = cmdArg[a+1];
			output.push_back(parsedTask);	
		}
		else if( cmdArg[a] == "-om" ) // matlab mat-file output
		{
			parsedTask.type = oMat;
			parsedTask.fileName = cmdArg[a+1];
			output.push_back(parsedTask);		
		}
		else // string not recognized - syntax error
		{
			return syntaxArg;
		}
	}
	
	if(input.size() != 1) return tooManyInput; // only one input allowed
	if(output.size() < 1 || output.size() > 2) return tooManyOutput; // only 1 or 2 outputs allowed
	
	/* a lot of the screening below could be removed if the processing part of convscp was 
	   improved. flexiblity in the lists of tasks makes it easy to add file types and even
	   enable allowing more combinations of inputs and outputs. here we *try* to screen 
	   for arguments combinations that we can't handle with the current code */
	
	for(list<task>::iterator outputIt = output.begin(), InputIt = input.begin(); // ensure an input isn't an output
		outputIt != output.end(); ++outputIt) {
		if(outputIt->type == InputIt->type) return errorMatch;
		if(outputIt->fileName == InputIt->fileName) return errorMatch;
	}
		
	if(output.size() > 1){ // ensure all outputs and their types are distinct
		for(list<task>::iterator outputIt1 = output.begin();
			outputIt1 != output.end(); ++outputIt1) {
			for(list<task>::iterator outputIt2 = output.begin(); 
				outputIt2 != outputIt1; ++outputIt2) {
					if(outputIt1->fileName == outputIt2->fileName) return errorMatch;
					if(outputIt1->type == outputIt2->type) return errorMatch;
			}
		}
	}
	
	return 0;
}

int parseError(int error)
{
	switch( error )
	{
	case numArg: 
		cerr << "Incorrect number of arguments. " << 
			"For help enter 'convscp --help'\n" << endl;
		break;
	case help: 
		usage();
		break;
	case tooManyInput:
		cerr << "Invalid number of input files. " << 
			"For help enter 'convscp --help'\n" << endl;
		break;
	case tooManyOutput:
		cerr << "Invalid number of output files. " << 
			"For help enter 'convscp --help'\n" << endl;
		break;
	case errorMatch:
		cerr << "Input and output file names and types must differ. " << 
			"For help enter 'convscp --help'\n" << endl;
		break;
	case nothingToDo: break;
	case syntaxArg:
		cerr << "Syntax error. " << 
			"For help enter 'convscp --help'\n" << endl;
	default: break;
	}

	return 0;
}
